@@ -10,7 +10,6 @@
 #include "src/base/numbers/cached-powers.h"
 #include "src/base/numbers/diy-fp.h"
 #include "src/base/numbers/double.h"
-#include "src/base/v8-fallthrough.h"
 
 namespace v8 {
 namespace base {
@@ -265,7 +264,7 @@ static inline void BiggestPowerTen(uint32_t number, int number_bits,
         *exponent = 9;
         break;
       }
-      V8_FALLTHROUGH;
+      [[fallthrough]];
     case 29:
     case 28:
     case 27:
@@ -274,7 +273,7 @@ static inline void BiggestPowerTen(uint32_t number, int number_bits,
         *exponent = 8;
         break;
       }
-      V8_FALLTHROUGH;
+      [[fallthrough]];
     case 26:
     case 25:
     case 24:
@@ -283,7 +282,7 @@ static inline void BiggestPowerTen(uint32_t number, int number_bits,
         *exponent = 7;
         break;
       }
-      V8_FALLTHROUGH;
+      [[fallthrough]];
     case 23:
     case 22:
     case 21:
@@ -293,7 +292,7 @@ static inline void BiggestPowerTen(uint32_t number, int number_bits,
         *exponent = 6;
         break;
       }
-      V8_FALLTHROUGH;
+      [[fallthrough]];
     case 19:
     case 18:
     case 17:
@@ -302,7 +301,7 @@ static inline void BiggestPowerTen(uint32_t number, int number_bits,
         *exponent = 5;
         break;
       }
-      V8_FALLTHROUGH;
+      [[fallthrough]];
     case 16:
     case 15:
     case 14:
@@ -311,7 +310,7 @@ static inline void BiggestPowerTen(uint32_t number, int number_bits,
         *exponent = 4;
         break;
       }
-      V8_FALLTHROUGH;
+      [[fallthrough]];
     case 13:
     case 12:
     case 11:
@@ -321,7 +320,7 @@ static inline void BiggestPowerTen(uint32_t number, int number_bits,
         *exponent = 3;
         break;
       }
-      V8_FALLTHROUGH;
+      [[fallthrough]];
     case 9:
     case 8:
     case 7:
@@ -330,7 +329,7 @@ static inline void BiggestPowerTen(uint32_t number, int number_bits,
         *exponent = 2;
         break;
       }
-      V8_FALLTHROUGH;
+      [[fallthrough]];
     case 6:
     case 5:
     case 4:
@@ -339,7 +338,7 @@ static inline void BiggestPowerTen(uint32_t number, int number_bits,
         *exponent = 1;
         break;
       }
-      V8_FALLTHROUGH;
+      [[fallthrough]];
     case 3:
     case 2:
     case 1:
@@ -348,7 +347,7 @@ static inline void BiggestPowerTen(uint32_t number, int number_bits,
         *exponent = 0;
         break;
       }
-      V8_FALLTHROUGH;
+      [[fallthrough]];
     case 0:
       *power = 0;
       *exponent = -1;
