@@ -6,23 +6,19 @@
 
 #include "src/compiler/js-heap-broker.h"
 #include "src/compiler/turboshaft/graph-builder.h"
+#include "src/compiler/turboshaft/phase.h"
 
 namespace v8::internal::compiler::turboshaft {
 
-base::Optional<BailoutReason> BuildGraphPhase::Run(PipelineData* data,
-                                                   Zone* temp_zone,
+base::Optional<BailoutReason> BuildGraphPhase::Run(Zone* temp_zone,
                                                    Linkage* linkage) {
-  Schedule* schedule = data->schedule();
-  data->reset_schedule();
+  Schedule* schedule = PipelineData::Get().schedule();
+  PipelineData::Get().reset_schedule();
   DCHECK_NOT_NULL(schedule);
-  data->CreateTurboshaftGraph();
 
-  UnparkedScopeIfNeeded scope(data->broker());
+  UnparkedScopeIfNeeded scope(PipelineData::Get().broker());
 
-  if (auto bailout = turboshaft::BuildGraph(
-          data->broker(), schedule, data->isolate(), data->graph_zone(),
-          temp_zone, &data->graph(), linkage, data->source_positions(),
-          data->node_origins())) {
+  if (auto bailout = turboshaft::BuildGraph(schedule, temp_zone, linkage)) {
     return bailout;
   }
   return {};
