@@ -207,7 +207,14 @@ class Worker : public std::enable_shared_from_this<Worker> {
 
   // Start running the given worker in another thread.
   static bool StartWorkerThread(Isolate* requester,
-                                std::shared_ptr<Worker> worker);
+                                std::shared_ptr<Worker> worker,
+                                base::Thread::Priority priority);
+
+  // Enters State::kTerminated for the Worker and resets the task runner.
+  void EnterTerminatedState();
+
+  // Returns the Worker instance for this thread.
+  static Worker* GetCurrentWorker();
 
  private:
   friend class ProcessMessageTask;
@@ -227,8 +234,9 @@ class Worker : public std::enable_shared_from_this<Worker> {
 
   class WorkerThread : public base::Thread {
    public:
-    explicit WorkerThread(std::shared_ptr<Worker> worker)
-        : base::Thread(base::Thread::Options("WorkerThread")),
+    explicit WorkerThread(std::shared_ptr<Worker> worker,
+                          base::Thread::Priority priority)
+        : base::Thread(base::Thread::Options("WorkerThread", priority)),
           worker_(std::move(worker)) {}
 
     void Run() override;
@@ -238,7 +246,9 @@ class Worker : public std::enable_shared_from_this<Worker> {
   };
 
   void ExecuteInThread();
-  static void PostMessageOut(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void PostMessageOut(const v8::FunctionCallbackInfo<v8::Value>& info);
+
+  static void SetCurrentWorker(Worker* worker);
 
   i::ParkingSemaphore out_semaphore_{0};
   SerializationDataQueue out_queue_;
@@ -262,7 +272,7 @@ class Worker : public std::enable_shared_from_this<Worker> {
   Isolate* isolate_ = nullptr;
 
   // Only accessed by the worker thread.
-  v8::Persistent<v8::Context> context_;
+  Global<Context> context_;
 };
 
 class PerIsolateData {
@@ -277,7 +287,7 @@ class PerIsolateData {
 
   class V8_NODISCARD RealmScope {
    public:
-    explicit RealmScope(PerIsolateData* data);
+    explicit RealmScope(Isolate* isolate, const Global<Context>& context);
     ~RealmScope();
 
    private:
@@ -343,7 +353,7 @@ class PerIsolateData {
   Global<FunctionTemplate> test_api_object_ctor_;
   Global<FunctionTemplate> dom_node_ctor_;
 
-  int RealmIndexOrThrow(const v8::FunctionCallbackInfo<v8::Value>& args,
+  int RealmIndexOrThrow(const v8::FunctionCallbackInfo<v8::Value>& info,
                         int arg_offset);
   int RealmFind(Local<Context> context);
 };
@@ -417,16 +427,16 @@ class ShellOptions {
   DisallowReassignment<bool> no_fail = {"no-fail", false};
   DisallowReassignment<bool> dump_counters = {"dump-counters", false};
   DisallowReassignment<bool> dump_counters_nvp = {"dump-counters-nvp", false};
+  DisallowReassignment<bool> dump_system_memory_stats = {
+      "dump-system-memory-stats", false};
   DisallowReassignment<bool> ignore_unhandled_promises = {
       "ignore-unhandled-promises", false};
   DisallowReassignment<bool> mock_arraybuffer_allocator = {
       "mock-arraybuffer-allocator", false};
   DisallowReassignment<size_t> mock_arraybuffer_allocator_limit = {
       "mock-arraybuffer-allocator-limit", 0};
-#if MULTI_MAPPED_ALLOCATOR_AVAILABLE
   DisallowReassignment<bool> multi_mapped_mock_allocator = {
       "multi-mapped-mock-allocator", false};
-#endif
   DisallowReassignment<bool> enable_inspector = {"enable-inspector", false};
   int num_isolates = 1;
   DisallowReassignment<v8::ScriptCompiler::CompileOptions, true>
@@ -448,6 +458,7 @@ class ShellOptions {
   DisallowReassignment<int> read_from_tcp_port = {"read-from-tcp-port", -1};
   DisallowReassignment<bool> enable_os_system = {"enable-os-system", false};
   DisallowReassignment<bool> quiet_load = {"quiet-load", false};
+  DisallowReassignment<bool> apply_priority = {"apply-priority", true};
   DisallowReassignment<int> thread_pool_size = {"thread-pool-size", 0};
   DisallowReassignment<bool> stress_delay_tasks = {"stress-delay-tasks", false};
   std::vector<const char*> arguments;
@@ -468,14 +479,6 @@ class ShellOptions {
   DisallowReassignment<bool> wasm_trap_handler = {"wasm-trap-handler", true};
 #endif  // V8_ENABLE_WEBASSEMBLY
   DisallowReassignment<bool> expose_fast_api = {"expose-fast-api", false};
-#if V8_ENABLE_SANDBOX
-  DisallowReassignment<bool> enable_sandbox_crash_filter = {
-      "enable-sandbox-crash-filter", false};
-#endif  // V8_ENABLE_SANDBOX
-  DisallowReassignment<bool> throw_on_failed_access_check = {
-      "throw-on-failed-access-check", false};
-  DisallowReassignment<bool> noop_on_failed_access_check = {
-      "noop-on-failed-access-check", false};
   DisallowReassignment<size_t> max_serializer_memory = {"max-serializer-memory",
                                                         1 * i::MB};
 };
@@ -494,14 +497,14 @@ class Shell : public i::AllStatic {
   enum class CodeType { kFileName, kString, kFunction, kInvalid, kNone };
 
   static bool ExecuteString(Isolate* isolate, Local<String> source,
-                            Local<String> name, PrintResult print_result,
+                            Local<String> name,
                             ReportExceptions report_exceptions,
-                            ProcessMessageQueue process_message_queue);
+                            Global<Value>* out_result = nullptr);
   static bool ExecuteModule(Isolate* isolate, const char* file_name);
   static bool LoadJSON(Isolate* isolate, const char* file_name);
   static void ReportException(Isolate* isolate, Local<Message> message,
                               Local<Value> exception);
-  static void ReportException(Isolate* isolate, TryCatch* try_catch);
+  static void ReportException(Isolate* isolate, const TryCatch& try_catch);
   static MaybeLocal<String> ReadFile(Isolate* isolate, const char* name,
                                      bool should_throw = true);
   static Local<String> WasmLoadSourceMapCallback(Isolate* isolate,
@@ -514,12 +517,10 @@ class Shell : public i::AllStatic {
   static void CollectGarbage(Isolate* isolate);
   static bool EmptyMessageQueues(Isolate* isolate);
   static bool CompleteMessageLoop(Isolate* isolate);
+  static bool FinishExecuting(Isolate* isolate, const Global<Context>& context);
 
   static bool HandleUnhandledPromiseRejections(Isolate* isolate);
 
-  static void PostForegroundTask(Isolate* isolate, std::unique_ptr<Task> task);
-  static void PostBlockingBackgroundTask(std::unique_ptr<Task> task);
-
   static std::unique_ptr<SerializationData> SerializeValue(
       Isolate* isolate, Local<Value> value, Local<Value> transfer);
   static MaybeLocal<Value> DeserializeValue(
@@ -534,58 +535,58 @@ class Shell : public i::AllStatic {
   static uint64_t GetTracingTimestampFromPerformanceTimestamp(
       double performance_timestamp);
 
-  static void PerformanceNow(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void PerformanceMark(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void PerformanceNow(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void PerformanceMark(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void PerformanceMeasure(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
   static void PerformanceMeasureMemory(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
 
-  static void RealmCurrent(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void RealmOwner(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void RealmGlobal(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void RealmCreate(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void RealmNavigate(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void RealmCurrent(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void RealmOwner(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void RealmGlobal(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void RealmCreate(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void RealmNavigate(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void RealmCreateAllowCrossRealmAccess(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
   static void RealmDetachGlobal(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void RealmDispose(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void RealmSwitch(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void RealmEval(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void RealmSharedGet(Local<String> property,
+      const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void RealmDispose(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void RealmSwitch(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void RealmEval(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void RealmSharedGet(Local<Name> property,
                              const PropertyCallbackInfo<Value>& info);
-  static void RealmSharedSet(Local<String> property, Local<Value> value,
+  static void RealmSharedSet(Local<Name> property, Local<Value> value,
                              const PropertyCallbackInfo<void>& info);
 
-  static void LogGetAndStop(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void LogGetAndStop(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void TestVerifySourcePositions(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
 
   static void InstallConditionalFeatures(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
 
   static void AsyncHooksCreateHook(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
   static void AsyncHooksExecutionAsyncId(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
   static void AsyncHooksTriggerAsyncId(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
 
-  static void SetPromiseHooks(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void SetPromiseHooks(const v8::FunctionCallbackInfo<v8::Value>& info);
 
-  static void EnableDebugger(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void DisableDebugger(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void EnableDebugger(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void DisableDebugger(const v8::FunctionCallbackInfo<v8::Value>& info);
 
   static void SerializerSerialize(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
   static void SerializerDeserialize(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
 
   static void ProfilerSetOnProfileEndListener(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
   static void ProfilerTriggerSample(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
 
   static bool HasOnProfileEndListener(Isolate* isolate);
 
@@ -594,45 +595,43 @@ class Shell : public i::AllStatic {
 
   static void ResetOnProfileEndListener(Isolate* isolate);
 
-  static void Print(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void PrintErr(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void WriteStdout(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void WaitUntilDone(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void NotifyDone(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void QuitOnce(v8::FunctionCallbackInfo<v8::Value>* args);
-  static void Quit(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void Terminate(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void Version(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void WriteFile(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void ReadFile(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void Print(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void PrintErr(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void WriteStdout(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void WaitUntilDone(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void NotifyDone(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void QuitOnce(v8::FunctionCallbackInfo<v8::Value>* info);
+  static void Quit(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void Terminate(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void Version(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void WriteFile(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void ReadFile(const v8::FunctionCallbackInfo<v8::Value>& info);
   static char* ReadChars(const char* name, int* size_out);
   static MaybeLocal<PrimitiveArray> ReadLines(Isolate* isolate,
                                               const char* name);
-  static void ReadBuffer(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void ReadBuffer(const v8::FunctionCallbackInfo<v8::Value>& info);
   static Local<String> ReadFromStdin(Isolate* isolate);
-  static void ReadLine(const v8::FunctionCallbackInfo<v8::Value>& args) {
-    args.GetReturnValue().Set(ReadFromStdin(args.GetIsolate()));
-  }
+  static void ReadLine(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void WriteChars(const char* name, uint8_t* buffer, size_t buffer_size);
-  static void ExecuteFile(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void SetTimeout(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void ExecuteFile(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void SetTimeout(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void ReadCodeTypeAndArguments(
-      const v8::FunctionCallbackInfo<v8::Value>& args, int index,
+      const v8::FunctionCallbackInfo<v8::Value>& info, int index,
       CodeType* code_type, Local<Value>* arguments = nullptr);
   static bool FunctionAndArgumentsToString(Local<Function> function,
                                            Local<Value> arguments,
                                            Local<String>* source,
                                            Isolate* isolate);
   static MaybeLocal<String> ReadSource(
-      const v8::FunctionCallbackInfo<v8::Value>& args, int index,
+      const v8::FunctionCallbackInfo<v8::Value>& info, int index,
       CodeType default_type);
-  static void WorkerNew(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void WorkerNew(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void WorkerPostMessage(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void WorkerGetMessage(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void WorkerTerminate(const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void WorkerGetMessage(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void WorkerTerminate(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void WorkerTerminateAndWait(
-      const v8::FunctionCallbackInfo<v8::Value>& args);
+      const v8::FunctionCallbackInfo<v8::Value>& info);
   // The OS object on the global object contains methods for performing
   // operating system calls:
   //
@@ -646,30 +645,39 @@ class Shell : public i::AllStatic {
   // milliseconds on the total running time of the program.  Exceptions are
   // thrown on timeouts or other errors or if the exit status of the program
   // indicates an error.
-  static void System(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void System(const v8::FunctionCallbackInfo<v8::Value>& info);
 
   // os.chdir(dir) changes directory to the given directory.  Throws an
   // exception/ on error.
-  static void ChangeDirectory(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void ChangeDirectory(const v8::FunctionCallbackInfo<v8::Value>& info);
 
   // os.setenv(variable, value) sets an environment variable.  Repeated calls to
   // this method leak memory due to the API of setenv in the standard C library.
-  static void SetEnvironment(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void UnsetEnvironment(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void SetEnvironment(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void UnsetEnvironment(const v8::FunctionCallbackInfo<v8::Value>& info);
 
   // os.umask(alue) calls the umask system call and returns the old umask.
-  static void SetUMask(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void SetUMask(const v8::FunctionCallbackInfo<v8::Value>& info);
 
   // os.mkdirp(name, mask) creates a directory.  The mask (if present) is anded
   // with the current umask.  Intermediate directories are created if necessary.
   // An exception is not thrown if the directory already exists.  Analogous to
   // the "mkdir -p" command.
-  static void MakeDirectory(const v8::FunctionCallbackInfo<v8::Value>& args);
-  static void RemoveDirectory(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void MakeDirectory(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void RemoveDirectory(const v8::FunctionCallbackInfo<v8::Value>& info);
+
+#ifdef V8_ENABLE_CONTINUATION_PRESERVED_EMBEDDER_DATA
+  static void GetContinuationPreservedEmbedderData(
+      const v8::FunctionCallbackInfo<v8::Value>& info);
+#endif  // V8_ENABLE_CONTINUATION_PRESERVER_EMBEDDER_DATA
+
+  static void GetExtrasBindingObject(
+      const v8::FunctionCallbackInfo<v8::Value>& info);
+
   static MaybeLocal<Promise> HostImportModuleDynamically(
       Local<Context> context, Local<Data> host_defined_options,
       Local<Value> resource_name, Local<String> specifier,
-      Local<FixedArray> import_assertions);
+      Local<FixedArray> import_attributes);
 
   static void ModuleResolutionSuccessCallback(
       const v8::FunctionCallbackInfo<v8::Value>& info);
@@ -682,7 +690,7 @@ class Shell : public i::AllStatic {
       Local<Context> initiator_context);
 
 #ifdef V8_FUZZILLI
-  static void Fuzzilli(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void Fuzzilli(const v8::FunctionCallbackInfo<v8::Value>& info);
 #endif  // V8_FUZZILLI
 
   // Data is of type DynamicImportData*. We use void* here to be able
@@ -760,10 +768,10 @@ class Shell : public i::AllStatic {
   static Counter* GetCounter(const char* name, bool is_histogram);
   static Local<String> Stringify(Isolate* isolate, Local<Value> value);
   static void RunShell(Isolate* isolate);
-  static bool RunMainIsolate(Isolate* isolate, bool last_run);
+  static bool RunMainIsolate(Isolate* isolate, bool keep_context_alive);
   static bool SetOptions(int argc, char* argv[]);
 
-  static void NodeTypeCallback(const v8::FunctionCallbackInfo<v8::Value>& args);
+  static void NodeTypeCallback(const v8::FunctionCallbackInfo<v8::Value>& info);
 
   static Local<FunctionTemplate> CreateEventTargetTemplate(Isolate* isolate);
   static Local<FunctionTemplate> CreateNodeTemplates(
@@ -781,9 +789,9 @@ class Shell : public i::AllStatic {
       Isolate* isolate);
 
   static MaybeLocal<Context> CreateRealm(
-      const v8::FunctionCallbackInfo<v8::Value>& args, int index,
+      const v8::FunctionCallbackInfo<v8::Value>& info, int index,
       v8::MaybeLocal<Value> global_object);
-  static void DisposeRealm(const v8::FunctionCallbackInfo<v8::Value>& args,
+  static void DisposeRealm(const v8::FunctionCallbackInfo<v8::Value>& info,
                            int index);
   static MaybeLocal<Module> FetchModuleTree(v8::Local<v8::Module> origin_module,
                                             v8::Local<v8::Context> context,
