@@ -1,8 +1,13 @@
 'use strict';
 
 const {
+  AggregateError,
+  AggregateErrorPrototype,
   Array,
+  ArrayBuffer,
+  ArrayBufferPrototype,
   ArrayIsArray,
+  ArrayPrototype,
   ArrayPrototypeFilter,
   ArrayPrototypeForEach,
   ArrayPrototypeIncludes,
@@ -13,14 +18,22 @@ const {
   ArrayPrototypePush,
   ArrayPrototypePushApply,
   ArrayPrototypeSlice,
-  ArrayPrototypeSplice,
   ArrayPrototypeSort,
+  ArrayPrototypeSplice,
   ArrayPrototypeUnshift,
   BigIntPrototypeValueOf,
+  Boolean,
+  BooleanPrototype,
   BooleanPrototypeValueOf,
+  DataView,
+  DataViewPrototype,
+  Date,
+  DatePrototype,
   DatePrototypeGetTime,
   DatePrototypeToISOString,
   DatePrototypeToString,
+  Error,
+  ErrorPrototype,
   ErrorPrototypeToString,
   Function,
   FunctionPrototype,
@@ -29,8 +42,10 @@ const {
   FunctionPrototypeSymbolHasInstance,
   FunctionPrototypeToString,
   JSONStringify,
-  MapPrototypeGetSize,
+  Map,
+  MapPrototype,
   MapPrototypeEntries,
+  MapPrototypeGetSize,
   MathFloor,
   MathMax,
   MathMin,
@@ -42,6 +57,7 @@ const {
   NumberIsNaN,
   NumberParseFloat,
   NumberParseInt,
+  NumberPrototype,
   NumberPrototypeToString,
   NumberPrototypeValueOf,
   Object,
@@ -58,21 +74,30 @@ const {
   ObjectPrototypePropertyIsEnumerable,
   ObjectSeal,
   ObjectSetPrototypeOf,
+  Promise,
+  PromisePrototype,
+  RangeError,
+  RangeErrorPrototype,
   ReflectApply,
   ReflectOwnKeys,
   RegExp,
+  RegExpPrototype,
   RegExpPrototypeExec,
   RegExpPrototypeSymbolReplace,
   RegExpPrototypeSymbolSplit,
   RegExpPrototypeToString,
-  SafeStringIterator,
   SafeMap,
   SafeSet,
+  SafeStringIterator,
+  Set,
+  SetPrototype,
   SetPrototypeGetSize,
   SetPrototypeValues,
   String,
+  StringPrototype,
   StringPrototypeCharCodeAt,
   StringPrototypeCodePointAt,
+  StringPrototypeEndsWith,
   StringPrototypeIncludes,
   StringPrototypeIndexOf,
   StringPrototypeLastIndexOf,
@@ -80,22 +105,30 @@ const {
   StringPrototypePadEnd,
   StringPrototypePadStart,
   StringPrototypeRepeat,
+  StringPrototypeReplace,
   StringPrototypeReplaceAll,
   StringPrototypeSlice,
   StringPrototypeSplit,
-  StringPrototypeEndsWith,
   StringPrototypeStartsWith,
   StringPrototypeToLowerCase,
   StringPrototypeTrim,
   StringPrototypeValueOf,
+  SymbolIterator,
   SymbolPrototypeToString,
   SymbolPrototypeValueOf,
-  SymbolIterator,
   SymbolToPrimitive,
   SymbolToStringTag,
+  TypeError,
+  TypeErrorPrototype,
+  TypedArray,
+  TypedArrayPrototype,
   TypedArrayPrototypeGetLength,
   TypedArrayPrototypeGetSymbolToStringTag,
   Uint8Array,
+  WeakMap,
+  WeakMapPrototype,
+  WeakSet,
+  WeakSetPrototype,
   globalThis,
   uncurryThis,
 } = primordials;
@@ -170,6 +203,11 @@ function pathToFileUrlHref(filepath) {
   return internalUrl.pathToFileURL(filepath).href;
 }
 
+function isURL(value) {
+  internalUrl ??= require('internal/url');
+  return typeof value.href === 'string' && value instanceof internalUrl.URL;
+}
+
 const builtInObjects = new SafeSet(
   ArrayPrototypeFilter(
     ObjectGetOwnPropertyNames(globalThis),
@@ -212,7 +250,6 @@ const keyStrRegExp = /^[a-zA-Z_][a-zA-Z_0-9]*$/;
 const numberRegExp = /^(0|[1-9][0-9]*)$/;
 
 const coreModuleRegExp = /^ {4}at (?:[^/\\(]+ \(|)node:(.+):\d+:\d+\)?$/;
-const nodeModulesRegExp = /[/\\]node_modules[/\\](.+?)(?=[/\\])/g;
 
 const classRegExp = /^(\s+[^(]*?)\s*{/;
 // eslint-disable-next-line node-core/no-unescaped-regexp-dot
@@ -593,16 +630,36 @@ function isInstanceof(object, proto) {
 }
 
 // Special-case for some builtin prototypes in case their `constructor` property has been tampered.
-const wellKnownPrototypes = new SafeMap();
-wellKnownPrototypes.set(ObjectPrototype, { name: 'Object', constructor: Object });
-wellKnownPrototypes.set(FunctionPrototype, { name: 'Function', constructor: Function });
+const wellKnownPrototypes = new SafeMap()
+  .set(ArrayPrototype, { name: 'Array', constructor: Array })
+  .set(ArrayBufferPrototype, { name: 'ArrayBuffer', constructor: ArrayBuffer })
+  .set(FunctionPrototype, { name: 'Function', constructor: Function })
+  .set(MapPrototype, { name: 'Map', constructor: Map })
+  .set(SetPrototype, { name: 'Set', constructor: Set })
+  .set(ObjectPrototype, { name: 'Object', constructor: Object })
+  .set(TypedArrayPrototype, { name: 'TypedArray', constructor: TypedArray })
+  .set(RegExpPrototype, { name: 'RegExp', constructor: RegExp })
+  .set(DatePrototype, { name: 'Date', constructor: Date })
+  .set(DataViewPrototype, { name: 'DataView', constructor: DataView })
+
+  .set(ErrorPrototype, { name: 'Error', constructor: Error })
+  .set(AggregateErrorPrototype, { name: 'AggregateError', constructor: AggregateError })
+  .set(RangeErrorPrototype, { name: 'RangeError', constructor: RangeError })
+  .set(TypeErrorPrototype, { name: 'TypeError', constructor: TypeError })
+
+  .set(BooleanPrototype, { name: 'Boolean', constructor: Boolean })
+  .set(NumberPrototype, { name: 'Number', constructor: Number })
+  .set(StringPrototype, { name: 'String', constructor: String })
+  .set(PromisePrototype, { name: 'Promise', constructor: Promise })
+  .set(WeakMapPrototype, { name: 'WeakMap', constructor: WeakMap })
+  .set(WeakSetPrototype, { name: 'WeakSet', constructor: WeakSet });
 
 function getConstructorName(obj, ctx, recurseTimes, protoProps) {
   let firstProto;
   const tmp = obj;
   while (obj || isUndetectableObject(obj)) {
     const wellKnownPrototypeNameAndConstructor = wellKnownPrototypes.get(obj);
-    if (wellKnownPrototypeNameAndConstructor != null) {
+    if (wellKnownPrototypeNameAndConstructor !== undefined) {
       const { name, constructor } = wellKnownPrototypeNameAndConstructor;
       if (FunctionPrototypeSymbolHasInstance(constructor, tmp)) {
         if (protoProps !== undefined && firstProto !== obj) {
@@ -714,6 +771,7 @@ function addPrototypeProperties(ctx, main, obj, recurseTimes, output) {
   } while (++depth !== 3);
 }
 
+/** @type {(constructor: string, tag: string, fallback: string, size?: string) => string} */
 function getPrefix(constructor, tag, fallback, size = '') {
   if (constructor === null) {
     if (tag !== '' && fallback !== tag) {
@@ -722,10 +780,20 @@ function getPrefix(constructor, tag, fallback, size = '') {
     return `[${fallback}${size}: null prototype] `;
   }
 
-  if (tag !== '' && constructor !== tag) {
-    return `${constructor}${size} [${tag}] `;
+  let result = `${constructor}${size} `;
+  if (tag !== '') {
+    const position = constructor.indexOf(tag);
+    if (position === -1) {
+      result += `[${tag}] `;
+    } else {
+      const endPos = position + tag.length;
+      if (endPos !== constructor.length &&
+        constructor[endPos] === constructor[endPos].toLowerCase()) {
+        result += `[${tag}] `;
+      }
+    }
   }
-  return `${constructor}${size} `;
+  return result;
 }
 
 // Look up the keys of the object.
@@ -820,12 +888,12 @@ function formatValue(ctx, value, recurseTimes, typedArray) {
         // Filter out the util module, its inspect function is special.
         maybeCustom !== inspect &&
         // Also filter out any prototype objects using the circular check.
-        !(value.constructor && value.constructor.prototype === value)) {
+        ObjectGetOwnPropertyDescriptor(value, 'constructor')?.value?.prototype !== value) {
       // This makes sure the recurseTimes are reported as before while using
       // a counter internally.
       const depth = ctx.depth === null ? null : ctx.depth - recurseTimes;
       const isCrossContext =
-        proxy !== undefined || !(context instanceof Object);
+        proxy !== undefined || !FunctionPrototypeSymbolHasInstance(Object, context);
       const ret = FunctionPrototypeCall(
         maybeCustom,
         context,
@@ -970,7 +1038,7 @@ function formatRaw(ctx, value, recurseTimes, typedArray) {
     keys = getKeys(value, ctx.showHidden);
     braces = ['{', '}'];
     if (typeof value === 'function') {
-      base = getFunctionBase(value, constructor, tag);
+      base = getFunctionBase(ctx, value, constructor, tag);
       if (keys.length === 0 && protoProps === undefined)
         return ctx.stylize(base, 'special');
     } else if (constructor === 'Object') {
@@ -1046,6 +1114,11 @@ function formatRaw(ctx, value, recurseTimes, typedArray) {
       if (keys.length === 0 && protoProps === undefined) {
         return base;
       }
+    } else if (isURL(value) && !(recurseTimes > ctx.depth && ctx.depth !== null)) {
+      base = value.href;
+      if (keys.length === 0 && protoProps === undefined) {
+        return base;
+      }
     } else {
       if (keys.length === 0 && protoProps === undefined) {
         if (isExternal(value)) {
@@ -1199,7 +1272,7 @@ function getClassBase(value, constructor, tag) {
   return `[${base}]`;
 }
 
-function getFunctionBase(value, constructor, tag) {
+function getFunctionBase(ctx, value, constructor, tag) {
   const stringified = FunctionPrototypeToString(value);
   if (StringPrototypeStartsWith(stringified, 'class') && stringified[stringified.length - 1] === '}') {
     const slice = StringPrototypeSlice(stringified, 5, -1);
@@ -1226,7 +1299,7 @@ function getFunctionBase(value, constructor, tag) {
   if (value.name === '') {
     base += ' (anonymous)';
   } else {
-    base += `: ${value.name}`;
+    base += `: ${typeof value.name === 'string' ? value.name : formatValue(ctx, value.name)}`;
   }
   base += ']';
   if (constructor !== type && constructor !== null) {
@@ -1252,17 +1325,137 @@ function identicalSequenceRange(a, b) {
           len++;
         }
         if (len > 3) {
-          return { len, offset: i };
+          return [len, i];
         }
       }
     }
   }
 
-  return { len: 0, offset: 0 };
+  return [0, 0];
 }
 
-function getStackString(error) {
-  return error.stack ? String(error.stack) : ErrorPrototypeToString(error);
+function getDuplicateErrorFrameRanges(frames) {
+  // Build a map: frame line -> sorted list of indices where it occurs
+  const result = [];
+  const lineToPositions = new SafeMap();
+
+  for (let i = 0; i < frames.length; i++) {
+    const positions = lineToPositions.get(frames[i]);
+    if (positions === undefined) {
+      lineToPositions.set(frames[i], [i]);
+    } else {
+      positions[positions.length] = i;
+    }
+  }
+
+  const minimumDuplicateRange = 3;
+  // Not enough duplicate lines to consider collapsing
+  if (frames.length - lineToPositions.size <= minimumDuplicateRange) {
+    return result;
+  }
+
+  for (let i = 0; i < frames.length - minimumDuplicateRange; i++) {
+    const positions = lineToPositions.get(frames[i]);
+    // Find the next occurrence of the same line after i, if any
+    if (positions.length === 1 || positions[positions.length - 1] === i) {
+      continue;
+    }
+
+    const current = positions.indexOf(i) + 1;
+    if (current === positions.length) {
+      continue;
+    }
+
+    // Theoretical maximum range, adjusted while iterating
+    let range = positions[positions.length - 1] - i;
+    if (range < minimumDuplicateRange) {
+      continue;
+    }
+    let extraSteps;
+    if (current + 1 < positions.length) {
+      // Optimize initial step size by choosing the greatest common divisor (GCD)
+      // of all candidate distances to the same frame line. This tends to match
+      // the true repeating block size and minimizes fallback iterations.
+      let gcdRange = 0;
+      for (let j = current; j < positions.length; j++) {
+        let distance = positions[j] - i;
+        while (distance !== 0) {
+          const remainder = gcdRange % distance;
+          if (gcdRange !== 0) {
+            // Add other possible ranges as fallback
+            extraSteps ??= new SafeSet();
+            extraSteps.add(gcdRange);
+          }
+          gcdRange = distance;
+          distance = remainder;
+        }
+        if (gcdRange === 1) break;
+      }
+      range = gcdRange;
+      if (extraSteps) {
+        extraSteps.delete(range);
+        extraSteps = [...extraSteps];
+      }
+    }
+    let maxRange = range;
+    let maxDuplicates = 0;
+
+    let duplicateRanges = 0;
+
+    for (let nextStart = i + range; /* ignored */ ; nextStart += range) {
+      let equalFrames = 0;
+      for (let j = 0; j < range; j++) {
+        if (frames[i + j] !== frames[nextStart + j]) {
+          break;
+        }
+        equalFrames++;
+      }
+      // Adjust the range to match different type of ranges.
+      if (equalFrames !== range) {
+        if (!extraSteps?.length) {
+          break;
+        }
+        // Memorize former range in case the smaller one would hide less.
+        if (duplicateRanges !== 0 && maxRange * maxDuplicates < range * duplicateRanges) {
+          maxRange = range;
+          maxDuplicates = duplicateRanges;
+        }
+        range = extraSteps.pop();
+        nextStart = i;
+        duplicateRanges = 0;
+        continue;
+      }
+      duplicateRanges++;
+    }
+
+    if (maxDuplicates !== 0 && maxRange * maxDuplicates >= range * duplicateRanges) {
+      range = maxRange;
+      duplicateRanges = maxDuplicates;
+    }
+
+    if (duplicateRanges * range >= 3) {
+      result.push(i + range, range, duplicateRanges);
+      // Skip over the collapsed portion to avoid overlapping matches.
+      i += range * (duplicateRanges + 1) - 1;
+    }
+  }
+
+  return result;
+}
+
+function getStackString(ctx, error) {
+  if (error.stack) {
+    if (typeof error.stack === 'string') {
+      return error.stack;
+    }
+    ctx.seen.push(error);
+    ctx.indentationLvl += 4;
+    const result = formatValue(ctx, error.stack);
+    ctx.indentationLvl -= 4;
+    ctx.seen.pop();
+    return `${ErrorPrototypeToString(error)}\n    ${result}`;
+  }
+  return ErrorPrototypeToString(error);
 }
 
 function getStackFrames(ctx, err, stack) {
@@ -1277,11 +1470,11 @@ function getStackFrames(ctx, err, stack) {
 
   // Remove stack frames identical to frames in cause.
   if (cause != null && isError(cause)) {
-    const causeStack = getStackString(cause);
+    const causeStack = getStackString(ctx, cause);
     const causeStackStart = StringPrototypeIndexOf(causeStack, '\n    at');
     if (causeStackStart !== -1) {
       const causeFrames = StringPrototypeSplit(StringPrototypeSlice(causeStack, causeStackStart + 1), '\n');
-      const { len, offset } = identicalSequenceRange(frames, causeFrames);
+      const { 0: len, 1: offset } = identicalSequenceRange(frames, causeFrames);
       if (len > 0) {
         const skipped = len - 2;
         const msg = `    ... ${skipped} lines matching cause stack trace ...`;
@@ -1289,14 +1482,42 @@ function getStackFrames(ctx, err, stack) {
       }
     }
   }
+
+  // Remove recursive repetitive stack frames in long stacks
+  if (frames.length > 10) {
+    const ranges = getDuplicateErrorFrameRanges(frames);
+
+    for (let i = ranges.length - 3; i >= 0; i -= 3) {
+      const offset = ranges[i];
+      const length = ranges[i + 1];
+      const duplicateRanges = ranges[i + 2];
+
+      const msg = `    ... collapsed ${length * duplicateRanges} duplicate lines ` +
+        'matching above ' +
+        (duplicateRanges > 1 ?
+          `${length} lines ${duplicateRanges} times...` :
+          'lines ...');
+      frames.splice(offset, length * duplicateRanges, ctx.stylize(msg, 'undefined'));
+    }
+  }
+
   return frames;
 }
 
+/** @type {(stack: string, constructor: string | null, name: unknown, tag: string) => string} */
 function improveStack(stack, constructor, name, tag) {
   // A stack trace may contain arbitrary data. Only manipulate the output
   // for "regular errors" (errors that "look normal") for now.
   let len = name.length;
 
+  if (typeof name !== 'string') {
+    stack = StringPrototypeReplace(
+      stack,
+      `${name}`,
+      `${name} [${StringPrototypeSlice(getPrefix(constructor, tag, 'Error'), 0, -1)}]`,
+    );
+  }
+
   if (constructor === null ||
       (StringPrototypeEndsWith(name, 'Error') &&
       StringPrototypeStartsWith(stack, name) &&
@@ -1305,9 +1526,9 @@ function improveStack(stack, constructor, name, tag) {
     if (constructor === null) {
       const start = RegExpPrototypeExec(/^([A-Z][a-z_ A-Z0-9[\]()-]+)(?::|\n {4}at)/, stack) ||
       RegExpPrototypeExec(/^([a-z_A-Z0-9-]*Error)$/, stack);
-      fallback = (start && start[1]) || '';
+      fallback = (start?.[1]) || '';
       len = fallback.length;
-      fallback = fallback || 'Error';
+      fallback ||= 'Error';
     }
     const prefix = StringPrototypeSlice(getPrefix(constructor, tag, fallback), 0, -1);
     if (name !== prefix) {
@@ -1329,8 +1550,8 @@ function removeDuplicateErrorKeys(ctx, keys, err, stack) {
   if (!ctx.showHidden && keys.length !== 0) {
     for (const name of ['name', 'message', 'stack']) {
       const index = ArrayPrototypeIndexOf(keys, name);
-      // Only hide the property in case it's part of the original stack
-      if (index !== -1 && StringPrototypeIncludes(stack, err[name])) {
+      // Only hide the property if it's a string and if it's part of the original stack
+      if (index !== -1 && (typeof err[name] !== 'string' || StringPrototypeIncludes(stack, err[name]))) {
         ArrayPrototypeSplice(keys, index, 1);
       }
     }
@@ -1339,16 +1560,45 @@ function removeDuplicateErrorKeys(ctx, keys, err, stack) {
 
 function markNodeModules(ctx, line) {
   let tempLine = '';
-  let nodeModule;
-  let pos = 0;
-  while ((nodeModule = nodeModulesRegExp.exec(line)) !== null) {
-    // '/node_modules/'.length === 14
-    tempLine += StringPrototypeSlice(line, pos, nodeModule.index + 14);
-    tempLine += ctx.stylize(nodeModule[1], 'module');
-    pos = nodeModule.index + nodeModule[0].length;
+  let lastPos = 0;
+  let searchFrom = 0;
+
+  while (true) {
+    const nodeModulePosition = StringPrototypeIndexOf(line, 'node_modules', searchFrom);
+    if (nodeModulePosition === -1) {
+      break;
+    }
+
+    // Ensure it's a path segment: must have a path separator before and after
+    const separator = line[nodeModulePosition - 1];
+    const after = line[nodeModulePosition + 12]; // 'node_modules'.length === 12
+
+    if ((after !== '/' && after !== '\\') || (separator !== '/' && separator !== '\\')) {
+      // Not a proper segment; continue searching
+      searchFrom = nodeModulePosition + 1;
+      continue;
+    }
+
+    const moduleStart = nodeModulePosition + 13; // Include trailing separator
+
+    // Append up to and including '/node_modules/'
+    tempLine += StringPrototypeSlice(line, lastPos, moduleStart);
+
+    let moduleEnd = StringPrototypeIndexOf(line, separator, moduleStart);
+    if (line[moduleStart] === '@') {
+      // Namespaced modules have an extra slash: @namespace/package
+      moduleEnd = StringPrototypeIndexOf(line, separator, moduleEnd + 1);
+    }
+
+    const nodeModule = StringPrototypeSlice(line, moduleStart, moduleEnd);
+    tempLine += ctx.stylize(nodeModule, 'module');
+
+    lastPos = moduleEnd;
+    searchFrom = moduleEnd;
   }
-  if (pos !== 0) {
-    line = tempLine + StringPrototypeSlice(line, pos);
+
+  if (lastPos !== 0) {
+    line = tempLine + StringPrototypeSlice(line, lastPos);
   }
   return line;
 }
@@ -1390,8 +1640,8 @@ function safeGetCWD() {
 }
 
 function formatError(err, constructor, tag, ctx, keys) {
-  const name = err.name != null ? String(err.name) : 'Error';
-  let stack = getStackString(err);
+  const name = err.name != null ? err.name : 'Error';
+  let stack = getStackString(ctx, err);
 
   removeDuplicateErrorKeys(ctx, keys, err, stack);
 
@@ -1611,23 +1861,28 @@ function formatNumber(fn, number, numericSeparator) {
     }
     return fn(`${number}`, 'number');
   }
+
+  const numberString = String(number);
   const integer = MathTrunc(number);
-  const string = String(integer);
+
   if (integer === number) {
-    if (!NumberIsFinite(number) || StringPrototypeIncludes(string, 'e')) {
-      return fn(string, 'number');
+    if (!NumberIsFinite(number) || StringPrototypeIncludes(numberString, 'e')) {
+      return fn(numberString, 'number');
     }
-    return fn(`${addNumericSeparator(string)}`, 'number');
+    return fn(addNumericSeparator(numberString), 'number');
   }
   if (NumberIsNaN(number)) {
-    return fn(string, 'number');
+    return fn(numberString, 'number');
   }
+
+  const decimalIndex = StringPrototypeIndexOf(numberString, '.');
+  const integerPart = StringPrototypeSlice(numberString, 0, decimalIndex);
+  const fractionalPart = StringPrototypeSlice(numberString, decimalIndex + 1);
+
   return fn(`${
-    addNumericSeparator(string)
+    addNumericSeparator(integerPart)
   }.${
-    addNumericSeparatorEnd(
-      StringPrototypeSlice(String(number), string.length + 1),
-    )
+    addNumericSeparatorEnd(fractionalPart)
   }`, 'number');
 }
 
@@ -1767,11 +2022,12 @@ function formatArray(ctx, value, recurseTimes) {
   const remaining = valLen - len;
   const output = [];
   for (let i = 0; i < len; i++) {
-    // Special handle sparse arrays.
-    if (!ObjectPrototypeHasOwnProperty(value, i)) {
+    const desc = ObjectGetOwnPropertyDescriptor(value, i);
+    if (desc === undefined) {
+      // Special handle sparse arrays.
       return formatSpecialArray(ctx, value, recurseTimes, len, output, i);
     }
-    ArrayPrototypePush(output, formatProperty(ctx, value, recurseTimes, i, kArrayType));
+    ArrayPrototypePush(output, formatProperty(ctx, value, recurseTimes, i, kArrayType, desc));
   }
   if (remaining > 0) {
     ArrayPrototypePush(output, remainingText(remaining));
@@ -1959,7 +2215,7 @@ function formatProperty(ctx, value, recurseTimes, key, type, desc,
                         original = value) {
   let name, str;
   let extra = ' ';
-  desc = desc || ObjectGetOwnPropertyDescriptor(value, key) ||
+  desc ||= ObjectGetOwnPropertyDescriptor(value, key) ||
     { value: value[key], enumerable: true };
   if (desc.value !== undefined) {
     const diff = (ctx.compact !== true || type !== kObjectType) ? 2 : 3;
