@@ -6,11 +6,19 @@
 #include "crypto/crypto_util.h"
 #include "env-inl.h"
 #include "memory_tracker-inl.h"
+#include "openssl/ec.h"
 #include "threadpoolwork-inl.h"
 #include "v8.h"
 
 namespace node {
 
+using ncrypto::BignumPointer;
+using ncrypto::ClearErrorOnReturn;
+using ncrypto::ECDSASigPointer;
+using ncrypto::ECKeyPointer;
+using ncrypto::EVPKeyCtxPointer;
+using ncrypto::EVPKeyPointer;
+using ncrypto::EVPMDCtxPointer;
 using v8::ArrayBuffer;
 using v8::BackingStore;
 using v8::Boolean;
@@ -20,8 +28,10 @@ using v8::HandleScope;
 using v8::Int32;
 using v8::Isolate;
 using v8::Just;
+using v8::JustVoid;
 using v8::Local;
 using v8::Maybe;
+using v8::MaybeLocal;
 using v8::Nothing;
 using v8::Object;
 using v8::Uint32;
@@ -31,19 +41,18 @@ namespace crypto {
 namespace {
 bool ValidateDSAParameters(EVP_PKEY* key) {
   /* Validate DSA2 parameters from FIPS 186-4 */
+  auto id = EVPKeyPointer::base_id(key);
 #if OPENSSL_VERSION_MAJOR >= 3
-  if (EVP_default_properties_is_fips_enabled(nullptr) &&
-      EVP_PKEY_DSA == EVP_PKEY_base_id(key)) {
+  if (EVP_default_properties_is_fips_enabled(nullptr) && EVP_PKEY_DSA == id) {
 #else
-  if (FIPS_mode() && EVP_PKEY_DSA == EVP_PKEY_base_id(key)) {
+  if (FIPS_mode() && EVP_PKEY_DSA == id) {
 #endif
     const DSA* dsa = EVP_PKEY_get0_DSA(key);
     const BIGNUM* p;
-    DSA_get0_pqg(dsa, &p, nullptr, nullptr);
-    size_t L = BN_num_bits(p);
     const BIGNUM* q;
-    DSA_get0_pqg(dsa, nullptr, &q, nullptr);
-    size_t N = BN_num_bits(q);
+    DSA_get0_pqg(dsa, &p, &q, nullptr);
+    int L = BignumPointer::GetBitCount(p);
+    int N = BignumPointer::GetBitCount(q);
 
     return (L == 1024 && N == 160) ||
            (L == 2048 && N == 224) ||
@@ -54,13 +63,12 @@ bool ValidateDSAParameters(EVP_PKEY* key) {
   return true;
 }
 
-bool ApplyRSAOptions(const ManagedEVPPKey& pkey,
+bool ApplyRSAOptions(const EVPKeyPointer& pkey,
                      EVP_PKEY_CTX* pkctx,
                      int padding,
                      const Maybe<int>& salt_len) {
-  if (EVP_PKEY_id(pkey.get()) == EVP_PKEY_RSA ||
-      EVP_PKEY_id(pkey.get()) == EVP_PKEY_RSA2 ||
-      EVP_PKEY_id(pkey.get()) == EVP_PKEY_RSA_PSS) {
+  int id = pkey.id();
+  if (id == EVP_PKEY_RSA || id == EVP_PKEY_RSA2 || id == EVP_PKEY_RSA_PSS) {
     if (EVP_PKEY_CTX_set_rsa_padding(pkctx, padding) <= 0)
       return false;
     if (padding == RSA_PKCS1_PSS_PADDING && salt_len.IsJust()) {
@@ -74,7 +82,7 @@ bool ApplyRSAOptions(const ManagedEVPPKey& pkey,
 
 std::unique_ptr<BackingStore> Node_SignFinal(Environment* env,
                                              EVPMDCtxPointer&& mdctx,
-                                             const ManagedEVPPKey& pkey,
+                                             const EVPKeyPointer& pkey,
                                              int padding,
                                              Maybe<int> pss_salt_len) {
   unsigned char m[EVP_MAX_MD_SIZE];
@@ -83,15 +91,13 @@ std::unique_ptr<BackingStore> Node_SignFinal(Environment* env,
   if (!EVP_DigestFinal_ex(mdctx.get(), m, &m_len))
     return nullptr;
 
-  int signed_sig_len = EVP_PKEY_size(pkey.get());
-  CHECK_GE(signed_sig_len, 0);
-  size_t sig_len = static_cast<size_t>(signed_sig_len);
+  size_t sig_len = pkey.size();
   std::unique_ptr<BackingStore> sig;
   {
     NoArrayBufferZeroFillScope no_zero_fill_scope(env->isolate_data());
     sig = ArrayBuffer::NewBackingStore(env->isolate(), sig_len);
   }
-  EVPKeyCtxPointer pkctx(EVP_PKEY_CTX_new(pkey.get(), nullptr));
+  EVPKeyCtxPointer pkctx = pkey.newCtx();
   if (pkctx && EVP_PKEY_sign_init(pkctx.get()) > 0 &&
       ApplyRSAOptions(pkey, pkctx.get(), padding, pss_salt_len) &&
       EVP_PKEY_CTX_set_signature_md(pkctx.get(), EVP_MD_CTX_md(mdctx.get())) >
@@ -107,9 +113,7 @@ std::unique_ptr<BackingStore> Node_SignFinal(Environment* env,
     } else if (sig_len != sig->ByteLength()) {
       std::unique_ptr<BackingStore> old_sig = std::move(sig);
       sig = ArrayBuffer::NewBackingStore(env->isolate(), sig_len);
-      memcpy(static_cast<char*>(sig->Data()),
-             static_cast<char*>(old_sig->Data()),
-             sig_len);
+      memcpy(sig->Data(), old_sig->Data(), sig_len);
     }
     return sig;
   }
@@ -117,22 +121,20 @@ std::unique_ptr<BackingStore> Node_SignFinal(Environment* env,
   return nullptr;
 }
 
-int GetDefaultSignPadding(const ManagedEVPPKey& m_pkey) {
-  return EVP_PKEY_id(m_pkey.get()) == EVP_PKEY_RSA_PSS ? RSA_PKCS1_PSS_PADDING :
-                                                         RSA_PKCS1_PADDING;
+int GetDefaultSignPadding(const EVPKeyPointer& m_pkey) {
+  return m_pkey.id() == EVP_PKEY_RSA_PSS ? RSA_PKCS1_PSS_PADDING
+                                         : RSA_PKCS1_PADDING;
 }
 
-unsigned int GetBytesOfRS(const ManagedEVPPKey& pkey) {
-  int bits, base_id = EVP_PKEY_base_id(pkey.get());
+unsigned int GetBytesOfRS(const EVPKeyPointer& pkey) {
+  int bits, base_id = pkey.base_id();
 
   if (base_id == EVP_PKEY_DSA) {
     const DSA* dsa_key = EVP_PKEY_get0_DSA(pkey.get());
     // Both r and s are computed mod q, so their width is limited by that of q.
-    bits = BN_num_bits(DSA_get0_q(dsa_key));
+    bits = BignumPointer::GetBitCount(DSA_get0_q(dsa_key));
   } else if (base_id == EVP_PKEY_EC) {
-    const EC_KEY* ec_key = EVP_PKEY_get0_EC_KEY(pkey.get());
-    const EC_GROUP* ec_group = EC_KEY_get0_group(ec_key);
-    bits = EC_GROUP_order_bits(ec_group);
+    bits = EC_GROUP_order_bits(ECKeyPointer::GetGroup(pkey));
   } else {
     return kNoDsaSignature;
   }
@@ -145,19 +147,23 @@ bool ExtractP1363(
     unsigned char* out,
     size_t len,
     size_t n) {
-  ECDSASigPointer asn1_sig(d2i_ECDSA_SIG(nullptr, &sig_data, len));
+  ncrypto::Buffer<const unsigned char> sig_buffer{
+      .data = sig_data,
+      .len = len,
+  };
+  auto asn1_sig = ECDSASigPointer::Parse(sig_buffer);
   if (!asn1_sig)
     return false;
 
-  const BIGNUM* pr = ECDSA_SIG_get0_r(asn1_sig.get());
-  const BIGNUM* ps = ECDSA_SIG_get0_s(asn1_sig.get());
-
-  return BN_bn2binpad(pr, out, n) > 0 && BN_bn2binpad(ps, out + n, n) > 0;
+  return BignumPointer::EncodePaddedInto(asn1_sig.r(), out, n) > 0 &&
+         BignumPointer::EncodePaddedInto(asn1_sig.s(), out + n, n) > 0;
 }
 
 // Returns the maximum size of each of the integers (r, s) of the DSA signature.
-std::unique_ptr<BackingStore> ConvertSignatureToP1363(Environment* env,
-    const ManagedEVPPKey& pkey, std::unique_ptr<BackingStore>&& signature) {
+std::unique_ptr<BackingStore> ConvertSignatureToP1363(
+    Environment* env,
+    const EVPKeyPointer& pkey,
+    std::unique_ptr<BackingStore>&& signature) {
   unsigned int n = GetBytesOfRS(pkey);
   if (n == kNoDsaSignature)
     return std::move(signature);
@@ -176,10 +182,9 @@ std::unique_ptr<BackingStore> ConvertSignatureToP1363(Environment* env,
 }
 
 // Returns the maximum size of each of the integers (r, s) of the DSA signature.
-ByteSource ConvertSignatureToP1363(
-    Environment* env,
-    const ManagedEVPPKey& pkey,
-    const ByteSource& signature) {
+ByteSource ConvertSignatureToP1363(Environment* env,
+                                   const EVPKeyPointer& pkey,
+                                   const ByteSource& signature) {
   unsigned int n = GetBytesOfRS(pkey);
   if (n == kNoDsaSignature)
     return ByteSource();
@@ -195,9 +200,7 @@ ByteSource ConvertSignatureToP1363(
   return std::move(out).release();
 }
 
-ByteSource ConvertSignatureToDER(
-      const ManagedEVPPKey& pkey,
-      ByteSource&& out) {
+ByteSource ConvertSignatureToDER(const EVPKeyPointer& pkey, ByteSource&& out) {
   unsigned int n = GetBytesOfRS(pkey);
   if (n == kNoDsaSignature)
     return std::move(out);
@@ -207,25 +210,19 @@ ByteSource ConvertSignatureToDER(
   if (out.size() != 2 * n)
     return ByteSource();
 
-  ECDSASigPointer asn1_sig(ECDSA_SIG_new());
+  auto asn1_sig = ECDSASigPointer::New();
   CHECK(asn1_sig);
-  BIGNUM* r = BN_new();
-  CHECK_NOT_NULL(r);
-  BIGNUM* s = BN_new();
-  CHECK_NOT_NULL(s);
-  CHECK_EQ(r, BN_bin2bn(sig_data, n, r));
-  CHECK_EQ(s, BN_bin2bn(sig_data + n, n, s));
-  CHECK_EQ(1, ECDSA_SIG_set0(asn1_sig.get(), r, s));
+  BignumPointer r(sig_data, n);
+  CHECK(r);
+  BignumPointer s(sig_data + n, n);
+  CHECK(s);
+  CHECK(asn1_sig.setParams(std::move(r), std::move(s)));
 
-  unsigned char* data = nullptr;
-  int len = i2d_ECDSA_SIG(asn1_sig.get(), &data);
+  auto buf = asn1_sig.encode();
+  if (buf.len <= 0) return ByteSource();
 
-  if (len <= 0)
-    return ByteSource();
-
-  CHECK_NOT_NULL(data);
-
-  return ByteSource::Allocated(data, len);
+  CHECK_NOT_NULL(buf.data);
+  return ByteSource::Allocated(buf);
 }
 
 void CheckThrow(Environment* env, SignBase::Error error) {
@@ -272,25 +269,13 @@ void CheckThrow(Environment* env, SignBase::Error error) {
   }
 }
 
-bool IsOneShot(const ManagedEVPPKey& key) {
-  switch (EVP_PKEY_id(key.get())) {
-    case EVP_PKEY_ED25519:
-    case EVP_PKEY_ED448:
-      return true;
-    default:
-      return false;
-  }
+bool IsOneShot(const EVPKeyPointer& key) {
+  return key.id() == EVP_PKEY_ED25519 || key.id() == EVP_PKEY_ED448;
 }
 
-bool UseP1363Encoding(const ManagedEVPPKey& key,
-                      const DSASigEnc& dsa_encoding) {
-  switch (EVP_PKEY_id(key.get())) {
-    case EVP_PKEY_EC:
-    case EVP_PKEY_DSA:
-      return dsa_encoding == kSigEncP1363;
-    default:
-      return false;
-  }
+bool UseP1363Encoding(const EVPKeyPointer& key, const DSASigEnc& dsa_encoding) {
+  return (key.id() == EVP_PKEY_EC || key.id() == EVP_PKEY_DSA) &&
+         dsa_encoding == kSigEncP1363;
 }
 }  // namespace
 
@@ -384,18 +369,17 @@ void Sign::SignUpdate(const FunctionCallbackInfo<Value>& args) {
   Decode<Sign>(args, [](Sign* sign, const FunctionCallbackInfo<Value>& args,
                         const char* data, size_t size) {
     Environment* env = Environment::GetCurrent(args);
-    if (UNLIKELY(size > INT_MAX))
+    if (size > INT_MAX) [[unlikely]]
       return THROW_ERR_OUT_OF_RANGE(env, "data is too long");
     Error err = sign->Update(data, size);
     crypto::CheckThrow(sign->env(), err);
   });
 }
 
-Sign::SignResult Sign::SignFinal(
-    const ManagedEVPPKey& pkey,
-    int padding,
-    const Maybe<int>& salt_len,
-    DSASigEnc dsa_sig_enc) {
+Sign::SignResult Sign::SignFinal(const EVPKeyPointer& pkey,
+                                 int padding,
+                                 const Maybe<int>& salt_len,
+                                 DSASigEnc dsa_sig_enc) {
   if (!mdctx_)
     return SignResult(kSignNotInitialised);
 
@@ -422,7 +406,10 @@ void Sign::SignFinal(const FunctionCallbackInfo<Value>& args) {
   ClearErrorOnReturn clear_error_on_return;
 
   unsigned int offset = 0;
-  ManagedEVPPKey key = ManagedEVPPKey::GetPrivateKeyFromJs(args, &offset, true);
+  auto data = KeyObjectData::GetPrivateKeyFromJs(args, &offset, true);
+  if (!data) [[unlikely]]
+    return;
+  const auto& key = data.GetAsymmetricKey();
   if (!key)
     return;
 
@@ -506,14 +493,14 @@ void Verify::VerifyUpdate(const FunctionCallbackInfo<Value>& args) {
                           const FunctionCallbackInfo<Value>& args,
                           const char* data, size_t size) {
     Environment* env = Environment::GetCurrent(args);
-    if (UNLIKELY(size > INT_MAX))
+    if (size > INT_MAX) [[unlikely]]
       return THROW_ERR_OUT_OF_RANGE(env, "data is too long");
     Error err = verify->Update(data, size);
     crypto::CheckThrow(verify->env(), err);
   });
 }
 
-SignBase::Error Verify::VerifyFinal(const ManagedEVPPKey& pkey,
+SignBase::Error Verify::VerifyFinal(const EVPKeyPointer& pkey,
                                     const ByteSource& sig,
                                     int padding,
                                     const Maybe<int>& saltlen,
@@ -529,7 +516,7 @@ SignBase::Error Verify::VerifyFinal(const ManagedEVPPKey& pkey,
   if (!EVP_DigestFinal_ex(mdctx.get(), m, &m_len))
     return kSignPublicKey;
 
-  EVPKeyCtxPointer pkctx(EVP_PKEY_CTX_new(pkey.get(), nullptr));
+  EVPKeyCtxPointer pkctx = pkey.newCtx();
   if (pkctx) {
     const int init_ret = EVP_PKEY_verify_init(pkctx.get());
     if (init_ret == -2) {
@@ -555,8 +542,9 @@ void Verify::VerifyFinal(const FunctionCallbackInfo<Value>& args) {
   ASSIGN_OR_RETURN_UNWRAP(&verify, args.This());
 
   unsigned int offset = 0;
-  ManagedEVPPKey pkey =
-      ManagedEVPPKey::GetPublicOrPrivateKeyFromJs(args, &offset);
+  auto data = KeyObjectData::GetPublicOrPrivateKeyFromJs(args, &offset);
+  if (!data) return;
+  const auto& pkey = data.GetAsymmetricKey();
   if (!pkey)
     return;
 
@@ -566,7 +554,7 @@ void Verify::VerifyFinal(const FunctionCallbackInfo<Value>& args) {
   }
 
   ArrayBufferOrViewContents<char> hbuf(args[offset]);
-  if (UNLIKELY(!hbuf.CheckSizeInt32()))
+  if (!hbuf.CheckSizeInt32()) [[unlikely]]
     return THROW_ERR_OUT_OF_RANGE(env, "buffer is too big");
 
   int padding = GetDefaultSignPadding(pkey);
@@ -627,7 +615,7 @@ void SignConfiguration::MemoryInfo(MemoryTracker* tracker) const {
   }
 }
 
-Maybe<bool> SignTraits::AdditionalConfig(
+Maybe<void> SignTraits::AdditionalConfig(
     CryptoJobMode mode,
     const FunctionCallbackInfo<Value>& args,
     unsigned int offset,
@@ -642,21 +630,22 @@ Maybe<bool> SignTraits::AdditionalConfig(
   params->mode =
       static_cast<SignConfiguration::Mode>(args[offset].As<Uint32>()->Value());
 
-  ManagedEVPPKey key;
   unsigned int keyParamOffset = offset + 1;
   if (params->mode == SignConfiguration::kVerify) {
-    key = ManagedEVPPKey::GetPublicOrPrivateKeyFromJs(args, &keyParamOffset);
+    auto data =
+        KeyObjectData::GetPublicOrPrivateKeyFromJs(args, &keyParamOffset);
+    if (!data) return Nothing<void>();
+    params->key = std::move(data);
   } else {
-    key = ManagedEVPPKey::GetPrivateKeyFromJs(args, &keyParamOffset, true);
+    auto data = KeyObjectData::GetPrivateKeyFromJs(args, &keyParamOffset, true);
+    if (!data) return Nothing<void>();
+    params->key = std::move(data);
   }
-  if (!key)
-    return Nothing<bool>();
-  params->key = key;
 
   ArrayBufferOrViewContents<char> data(args[offset + 5]);
-  if (UNLIKELY(!data.CheckSizeInt32())) {
+  if (!data.CheckSizeInt32()) [[unlikely]] {
     THROW_ERR_OUT_OF_RANGE(env, "data is too big");
-    return Nothing<bool>();
+    return Nothing<void>();
   }
   params->data = mode == kCryptoJobAsync
       ? data.ToCopy()
@@ -667,7 +656,7 @@ Maybe<bool> SignTraits::AdditionalConfig(
     params->digest = EVP_get_digestbyname(*digest);
     if (params->digest == nullptr) {
       THROW_ERR_CRYPTO_INVALID_DIGEST(env, "Invalid digest: %s", *digest);
-      return Nothing<bool>();
+      return Nothing<void>();
     }
   }
 
@@ -686,23 +675,22 @@ Maybe<bool> SignTraits::AdditionalConfig(
     if (params->dsa_encoding != kSigEncDER &&
         params->dsa_encoding != kSigEncP1363) {
       THROW_ERR_OUT_OF_RANGE(env, "invalid signature encoding");
-      return Nothing<bool>();
+      return Nothing<void>();
     }
   }
 
   if (params->mode == SignConfiguration::kVerify) {
     ArrayBufferOrViewContents<char> signature(args[offset + 10]);
-    if (UNLIKELY(!signature.CheckSizeInt32())) {
+    if (!signature.CheckSizeInt32()) [[unlikely]] {
       THROW_ERR_OUT_OF_RANGE(env, "signature is too big");
-      return Nothing<bool>();
+      return Nothing<void>();
     }
     // If this is an EC key (assuming ECDSA) we need to convert the
     // the signature from WebCrypto format into DER format...
-    ManagedEVPPKey m_pkey = params->key;
-    Mutex::ScopedLock lock(*m_pkey.mutex());
-    if (UseP1363Encoding(m_pkey, params->dsa_encoding)) {
-      params->signature =
-          ConvertSignatureToDER(m_pkey, signature.ToByteSource());
+    Mutex::ScopedLock lock(params->key.mutex());
+    const auto& akey = params->key.GetAsymmetricKey();
+    if (UseP1363Encoding(akey, params->dsa_encoding)) {
+      params->signature = ConvertSignatureToDER(akey, signature.ToByteSource());
     } else {
       params->signature = mode == kCryptoJobAsync
           ? signature.ToCopy()
@@ -710,7 +698,7 @@ Maybe<bool> SignTraits::AdditionalConfig(
     }
   }
 
-  return Just(true);
+  return JustVoid();
 }
 
 bool SignTraits::DeriveBits(Environment* env,
@@ -721,25 +709,19 @@ bool SignTraits::DeriveBits(Environment* env,
   EVPMDCtxPointer context(EVP_MD_CTX_new());
   EVP_PKEY_CTX* ctx = nullptr;
 
+  const auto& key = params.key.GetAsymmetricKey();
+
   switch (params.mode) {
     case SignConfiguration::kSign:
       if (!EVP_DigestSignInit(
-              context.get(),
-              &ctx,
-              params.digest,
-              nullptr,
-              params.key.get())) {
+              context.get(), &ctx, params.digest, nullptr, key.get())) {
         if (can_throw) crypto::CheckThrow(env, SignBase::Error::kSignInit);
         return false;
       }
       break;
     case SignConfiguration::kVerify:
       if (!EVP_DigestVerifyInit(
-              context.get(),
-              &ctx,
-              params.digest,
-              nullptr,
-              params.key.get())) {
+              context.get(), &ctx, params.digest, nullptr, key.get())) {
         if (can_throw) crypto::CheckThrow(env, SignBase::Error::kSignInit);
         return false;
       }
@@ -747,24 +729,20 @@ bool SignTraits::DeriveBits(Environment* env,
   }
 
   int padding = params.flags & SignConfiguration::kHasPadding
-      ? params.padding
-      : GetDefaultSignPadding(params.key);
+                    ? params.padding
+                    : GetDefaultSignPadding(key);
 
   Maybe<int> salt_length = params.flags & SignConfiguration::kHasSaltLength
       ? Just<int>(params.salt_length) : Nothing<int>();
 
-  if (!ApplyRSAOptions(
-          params.key,
-          ctx,
-          padding,
-          salt_length)) {
+  if (!ApplyRSAOptions(key, ctx, padding, salt_length)) {
     if (can_throw) crypto::CheckThrow(env, SignBase::Error::kSignPrivateKey);
     return false;
   }
 
   switch (params.mode) {
     case SignConfiguration::kSign: {
-      if (IsOneShot(params.key)) {
+      if (IsOneShot(key)) {
         size_t len;
         if (!EVP_DigestSign(
             context.get(),
@@ -782,8 +760,7 @@ bool SignTraits::DeriveBits(Environment* env,
                             &len,
                             params.data.data<unsigned char>(),
                             params.data.size())) {
-          if (can_throw)
-            crypto::CheckThrow(env, SignBase::Error::kSignPrivateKey);
+          crypto::CheckThrow(env, SignBase::Error::kSignPrivateKey);
           return false;
         }
         *out = std::move(buf).release(len);
@@ -806,9 +783,8 @@ bool SignTraits::DeriveBits(Environment* env,
           return false;
         }
 
-        if (UseP1363Encoding(params.key, params.dsa_encoding)) {
-          *out = ConvertSignatureToP1363(
-              env, params.key, std::move(buf).release());
+        if (UseP1363Encoding(key, params.dsa_encoding)) {
+          *out = ConvertSignatureToP1363(env, key, std::move(buf).release());
         } else {
           *out = std::move(buf).release(len);
         }
@@ -833,22 +809,16 @@ bool SignTraits::DeriveBits(Environment* env,
   return true;
 }
 
-Maybe<bool> SignTraits::EncodeOutput(
-    Environment* env,
-    const SignConfiguration& params,
-    ByteSource* out,
-    Local<Value>* result) {
+MaybeLocal<Value> SignTraits::EncodeOutput(Environment* env,
+                                           const SignConfiguration& params,
+                                           ByteSource* out) {
   switch (params.mode) {
     case SignConfiguration::kSign:
-      *result = out->ToArrayBuffer(env);
-      break;
+      return out->ToArrayBuffer(env);
     case SignConfiguration::kVerify:
-      *result = Boolean::New(env->isolate(), out->data<char>()[0] == 1);
-      break;
-    default:
-      UNREACHABLE();
+      return Boolean::New(env->isolate(), out->data<char>()[0] == 1);
   }
-  return Just(!result->IsEmpty());
+  UNREACHABLE();
 }
 
 }  // namespace crypto
